package lib;

import lib.Vec.Vec3;

/**
	Represents a Point in a 3D `Int` System.
	Basicly a wrapper for Vec3<Int> with some extra functions.
	`Y` represents the height of the point.
**/
@:forward(x, y, z)
abstract BlockPos(Vec3<Int>) from Vec3<Int> to Vec3<Int> {
	public inline function new(x:Int, y:Int, z:Int) {
		this = new Vec3(x, y, z);
	}

	@:op(A + B)
	public function add(rhs:BlockPos):BlockPos {
		return {
			y: this.y + rhs.y,
			x: this.x + rhs.x,
			z: this.z + rhs.z
		};
	}

	@:op(A - B)
	public function sub(rhs:BlockPos):BlockPos {
		return {
			y: this.y - rhs.y,
			x: this.x - rhs.x,
			z: this.z - rhs.z
		};
	}

	@:op(A * B)
	public function multiplyScalar(rhs:Int):BlockPos {
		return {
			y: this.y * rhs,
			x: this.x * rhs,
			z: this.z * rhs
		};
	}

	@:op(-A)
	public function negate():BlockPos {
		return {
			y: -this.y,
			x: -this.x,
			z: -this.z
		};
	}

	@:op(A == B)
	public function equals(rhs:BlockPos):Bool {
		return this.x == rhs.x && this.y == rhs.y && this.z == rhs.z;
	}

	@:op(A != B)
	public function notEquals(rhs:BlockPos):Bool {
		return !equals(rhs);
	}

	public function dot(rhs:BlockPos):Float {
		return this.x * rhs.x + this.y * rhs.y + this.z * rhs.z;
	}

	public function cross(rhs:BlockPos):BlockPos {
		return {
			x: this.x * rhs.y - this.y * rhs.x,
			y: this.y * rhs.z - this.z * rhs.y,
			z: this.z * rhs.x - this.x * rhs.z,
		};
	}

	public function length():Float {
		return Math.sqrt(this.x * this.x + this.y * this.y + this.z * this.z);
	}

	public function hashCode():Int {
		return (this.x * 73856093) ^ (this.y * 19349663) ^ (this.z * 83492791);
	}

	public function toString():String {
		return 'BlockPos(${this.x}, ${this.y}, ${this.z})';
	}

	public function distance(rhs:BlockPos):Float {
		return Math.sqrt(Math.pow(this.x - rhs.x, 2) + Math.pow(this.y - rhs.y, 2) + Math.pow(this.z - rhs.z, 2));
	}

	/**
		Returns a list of positions neighboring the block. No Diagonal.
	**/
	public function neighbors():Array<BlockPos> {
		return [
			new BlockPos(this.x + 1, this.y + 0, this.z + 0), // Front
			new BlockPos(this.x - 1, this.y + 0, this.z + 0), // Back
			new BlockPos(this.x + 0, this.y + 0, this.z - 1), // Left
			new BlockPos(this.x + 0, this.y + 0, this.z + 1), // Right
			new BlockPos(this.x + 0, this.y - 1, this.z + 0), // Bot
			new BlockPos(this.x + 0, this.y + 1, this.z + 0), // Top
		];
	}
}
